/*
    libporty - async background processes for windows
    Copyright (C) 2010..2012  Tibor 'Igor2' Palinkas

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    Project page: http://repo.hu/projects/libporty
    Author: libporty@igor2.repo.hu
*/

#include <stdio.h>
#include <libporty/bgtask/bgtask.h>

static void P_net_async_bgproc_set_inheritance(P_net_socket *inherit_list, int inherit_len)
{
	int n;

	for(n = 0; n < 1024; n++)
		SetHandleInformation(n, HANDLE_FLAG_INHERIT, 0);

	SetHandleInformation(GetStdHandle(STD_INPUT_HANDLE), HANDLE_FLAG_INHERIT, HANDLE_FLAG_INHERIT);
	SetHandleInformation(GetStdHandle(STD_OUTPUT_HANDLE), HANDLE_FLAG_INHERIT, HANDLE_FLAG_INHERIT);
	SetHandleInformation(GetStdHandle(STD_ERROR_HANDLE), HANDLE_FLAG_INHERIT, HANDLE_FLAG_INHERIT);

	for(n = 0; n < inherit_len; n++)
		SetHandleInformation(inherit_list[n], HANDLE_FLAG_INHERIT, HANDLE_FLAG_INHERIT);
}


int P_net_async_bgproc_inherit(P_net_socket *fd, char *argv[], P_net_socket *ctrlfd, P_net_socket *fd_err, P_net_socket *inherit_list, int inherit_len)
{
	P_net_socket chfd, chfd_err;
	int ret, argc, n, len, innerpid;
	PROCESS_INFORMATION piProcInfo;
	STARTUPINFO siStartInfo;
	char *cmdline, *end;

	P_net_async_bgproc_set_inheritance(inherit_list, inherit_len);

	if (fd != NULL) {
		ret = P_net_local_socket(fd, &chfd, P_net_nonblock_full);
		if (ret != 0)
			return ret;
	}
	if (fd_err != NULL) {
		ret = P_net_local_socket(fd_err, &chfd_err, P_net_nonblock_full);
		if (ret != 0)
			return ret;
	}

	for(argc = 0, len = 0; argv[argc] != NULL; argc++)
		len += strlen(argv[argc]) + 2;

#warning this should be caluclated according to expected length of socket being printed
	len += inherit_len * 16;

	cmdline = malloc(len + 32);

	end = cmdline + sprintf(cmdline, "%s %s", P_SHELL, P_SHELL_QUOTE);

	for(n = 0; n < argc; n++)
		end += sprintf(end, "%s ", argv[n]);
	end += sprintf(end, "%s", P_SHELL_QUOTE);

	end--;
	*end = '\0';

/* Set up members of the PROCESS_INFORMATION structure. */
	ZeroMemory( &piProcInfo, sizeof(PROCESS_INFORMATION) );

/* Set up members of the STARTUPINFO structure. */
	ZeroMemory( &siStartInfo, sizeof(STARTUPINFO) );
	siStartInfo.cb         = sizeof(STARTUPINFO);

	P_net_set_blocking(chfd);
	if (fd != NULL) {
		siStartInfo.hStdInput  = (HANDLE *)chfd;
		siStartInfo.hStdOutput = (HANDLE *)chfd;
	}
	else {
		siStartInfo.hStdInput  = GetStdHandle(STD_INPUT_HANDLE);
		siStartInfo.hStdOutput = GetStdHandle(STD_OUTPUT_HANDLE);
	}

	if (fd_err != NULL)
		siStartInfo.hStdError  = (HANDLE *)chfd_err;
	else
		siStartInfo.hStdError  = GetStdHandle(STD_ERROR_HANDLE);

	siStartInfo.dwFlags   |= STARTF_USESTDHANDLES;

/* Create the child process. */
	if (CreateProcess(NULL,
		cmdline,          /* command line */
		NULL,          /* process security attributes */
		NULL,          /* primary thread security attributes */
		TRUE,          /* handles are inherited */
		0,             /* creation flags */
		NULL,          /* use parent's environment */
		NULL,          /* use parent's current directory */
		&siStartInfo,  /* STARTUPINFO pointer */
		&piProcInfo)   /* receives PROCESS_INFORMATION */
		== 0) {
		return -128;
	}

	if (ctrlfd != NULL) {
		int cfd;

		ret = P_net_local_socket(ctrlfd, &cfd, P_net_nonblock_full);
		if (ret != 0)
			return ret;

		watchdog_add(&cfd, &(piProcInfo.hProcess), &chfd);
	}

	return 0;
}

int P_net_async_bgproc(P_net_socket *fd, char *argv[], P_net_socket *ctrlfd)
{
	return P_net_async_bgproc_inherit(fd, argv, ctrlfd, NULL, NULL, 0);
}

#define STATUS_KILL 0xCC000001
#define STATUS_TERM 0xCC000002

static int watchdog_wait(P_net_socket cfd, int numpids, P_pid_t *pids_, bgrpoc_status_t *exit_status)
{
	int err;
	int *pids = pids_;
	int n, w;

	for(n = 0; n < numpids; n += MAXIMUM_WAIT_OBJECTS) {
		w = numpids - n;
		if (w > MAXIMUM_WAIT_OBJECTS)
			w = MAXIMUM_WAIT_OBJECTS;
/*		printf("watchdog waiting for %d pids starting from %d (all pids=%d), first pid: %d\n", w, n, numpids, (int)pids[n]);*/
		err = WaitForMultipleObjects(w, pids + n, 0, 0);
		if (err == WAIT_TIMEOUT) {
/*			printf(" timeout\n");*/
			return -1;
		}

		if ((err >= 0) && (err < WAIT_ABANDONED_0)) {
			err -= WAIT_OBJECT_0;
			GetExitCodeProcess(pids[err], &(exit_status->code));
			if ((exit_status->code == STATUS_TERM) || (exit_status->code == STATUS_KILL))
				exit_status->reason = BGPROC_KILLED;
			else if ((exit_status->code & 0xC0000000) == 0xC0000000)
				exit_status->reason = BGPROC_ABNORMAL;
			else
				exit_status->reason = BGPROC_EXITED;
/*			printf("watchdog EXITED: entry=%d pid=%d status=%d\n", err, pids[err], *exit_status);*/
			return err;
		}
/*
		else
			printf("  is %d\n", GetLastError());
*/
	}
	return -1;
}

static void watchdog_kill(P_pid_t pid, int hard)
{
	TerminateProcess(pid, hard ? STATUS_KILL : STATUS_TERM);
}
